theory cnf_com imports executions begin
  
(* conflict *)
fun cnf :: "witnessed_execution \<Rightarrow> event rel" 
  where "cnf (X,rf,co) = (((W X \<times> W X) \<union> (W X \<times> R X) \<union> (R X \<times> W X)) \<inter> sloc X) - Id"
    
(* extended communication *)
fun ecom :: "witnessed_execution \<Rightarrow> event rel" 
  where "ecom (X,rf,co) = rf \<union> co \<union> fr(X,rf,co) \<union> (co O rf)"
    
lemma trans_implies_rtrancl_is_reflcl: 
  "trans r \<Longrightarrow> r\<^sup>* = r\<^sup>="
  using trancl_id trancl_reflcl trans_reflclI 
  by blast
  
lemma sym_sloc:
  "wf_exec(X,rf,co) \<Longrightarrow> (e1,e2) \<in> sloc X \<Longrightarrow> (e2,e1) \<in> sloc X"
  apply (erule wf_exec.cases)
  apply (elim equivE)
  apply (rule symD)
   apply auto
  done

lemma sloc_writes_in_com:
  "wf_exec(X,rf,co) \<Longrightarrow> e1 \<in> W X \<Longrightarrow> e2 \<in> W X \<Longrightarrow> 
(e1,e2) \<in> sloc X \<Longrightarrow> e1 \<noteq> e2 \<Longrightarrow> 
(e1, e2) \<in> co \<or> (e2, e1) \<in> co"
  apply (erule wf_exec.cases)
  apply auto
  done
    
lemma sloc_read_write_without_rf_in_com:
  "wf_exec(X,rf,co) \<Longrightarrow> e1 \<in> W X \<Longrightarrow> e2 \<in> R X \<Longrightarrow> 
(e1,e2) \<in> sloc X \<Longrightarrow> e1 \<noteq> e2 \<Longrightarrow> e2 \<notin> Range rf \<Longrightarrow> 
(e2, e1) \<in> fr (X,rf,co)"
  apply (erule wf_exec.cases)
  apply clarsimp
  apply (intro conjI)
   apply (elim equivE)
   apply (rule symD, assumption+)
  apply clarsimp
  apply auto
  done
   
lemma sloc_read_write_with_rf_in_com:
  "wf_exec(X,rf,co) \<Longrightarrow> e1 \<in> W X \<Longrightarrow> e2 \<in> R X \<Longrightarrow> 
(e1,e2) \<in> sloc X \<Longrightarrow> e1 \<noteq> e2 \<Longrightarrow> e2 \<in> Range rf \<Longrightarrow> 
(e1, e2) \<in> rf \<or> (e1, e2) \<in> co O rf \<or> (e2, e1) \<in> fr (X,rf,co)"
  apply (erule wf_exec.cases)
  apply clarsimp
  apply (erule impE)
   apply (elim equivE)
   apply (rule symD, assumption+)
  apply (elim relcompEpair)
  apply (subgoal_tac "(b, e1) \<in> (co\<inverse>)\<^sup>=")
   apply clarsimp 
   apply (elim disjE)
    apply (intro relcompI)
     apply assumption+
   apply (intro relcompI)
    apply simp+
  apply (simp add: trans_implies_rtrancl_is_reflcl)
  done
   
theorem cnf_com:
  assumes "wf_exec(X,rf,co)"
  shows "cnf (X,rf,co) \<subseteq> ecom (X,rf,co) \<union> (ecom (X,rf,co))\<inverse>"
  apply (intro subsetI)
  apply (case_tac x)
  apply (simp del: cnf.simps ecom.simps)  
proof -
  fix e1 e2
  assume cnf: "(e1, e2) \<in> cnf (X, rf, co)"
  hence cnf': "(e2, e1) \<in> cnf (X, rf, co)" 
    using sym_sloc[OF assms] by auto
  {
    assume "e1 \<in> W X" and "e2 \<in> W X"
    hence "(e1, e2) \<in> co \<or> (e2, e1) \<in> co"
      using sloc_writes_in_com[OF assms] and cnf 
      by auto
  } 
  moreover {
    assume "e1 \<in> W X" and "e2 \<in> R X" and "e2 \<notin> Range rf"
    hence "(e2, e1) \<in> fr (X,rf,co)" 
      using sloc_read_write_without_rf_in_com[OF assms] and cnf
      by auto
  } 
  moreover {
    assume "e1 \<in> W X" and "e2 \<in> R X" and "e2 \<in> Range rf"
    hence "(e1, e2) \<in> rf \<or> (e1, e2) \<in> co O rf \<or> (e2, e1) \<in> fr (X,rf,co)"
      using sloc_read_write_with_rf_in_com[OF assms] and cnf
      by auto
  } 
  moreover {
    assume "e2 \<in> W X" and "e1 \<in> R X" and "e1 \<notin> Range rf"
    hence "(e1, e2) \<in> fr (X,rf,co)" 
      using sloc_read_write_without_rf_in_com[OF assms] and cnf
      by auto
  } 
  moreover {
    assume "e2 \<in> W X" and "e1 \<in> R X" and "e1 \<in> Range rf"
    hence "(e2, e1) \<in> rf \<or> (e2, e1) \<in> co O rf \<or> (e1, e2) \<in> fr (X,rf,co)"
      using sloc_read_write_with_rf_in_com[OF assms] and cnf'
      by auto
  } 
  ultimately 
  show "(e1, e2) \<in> ecom (X, rf, co) \<or> (e2, e1) \<in> ecom (X, rf, co)" 
    using cnf by auto
qed
  
theorem com_cnf:
  assumes "wf_exec(X,rf,co)"
  shows "ecom (X,rf,co) \<union> (ecom (X,rf,co))\<inverse> \<subseteq> cnf (X,rf,co)"
  using assms
  apply (intro subsetI)
  apply (cases rule: wf_exec.cases, assumption)
  apply clarsimp
  apply (elim disjE)
         apply (intro conjI)
           apply (rule disjI2, rule disjI1)
           apply auto[3]
        apply (intro conjI)
          apply (rule disjI1)
          apply auto[3]
       apply (intro conjI)
         apply auto[3]
      apply (intro conjI)
        apply (rule disjI2, rule disjI1)
        apply (elim relcompE)
        apply auto[1]
       apply (elim relcompE)
       apply (erule equivE[where r="sloc X"]) 
       apply (erule transE[where r="sloc X"])
         defer defer
         apply assumption
        apply (elim relcompE)
        apply (subgoal_tac "x \<in> W X")
        apply auto[2]
       apply (intro conjI)
         apply (rule disjI2, rule disjI2)
         apply auto[1]
        apply (rule sym_sloc, assumption)
        apply auto[2]
      apply (intro conjI)
        apply (rule disjI1)
        apply auto[1]
       apply (rule sym_sloc, assumption)
       apply auto[2]
     apply (intro conjI)
       apply (rule disjI2, rule disjI1)
       apply auto[1]
      apply (rule sym_sloc, assumption)
      apply auto[2]
    apply (intro conjI)
      apply (rule disjI2, rule disjI2)
      apply (elim relcompE)
      apply auto[1]
     apply (elim relcompE)
     apply (erule equivE[where r="sloc X"]) 
     apply (erule transE[where r="sloc X"])
       defer defer
       apply assumption
      apply (elim relcompE)
      apply (subgoal_tac "x \<in> W X")
       apply auto[4]
   apply (subgoal_tac "(a, y) \<in> sloc X")
    apply assumption
   apply (rule sym_sloc, assumption)
   apply auto[1]
  apply (rule sym_sloc, assumption)
  apply auto[1]
  done
    
corollary com_eq_cnf:
  assumes "wf_exec(X,rf,co)"
  shows "ecom (X,rf,co) \<union> (ecom (X,rf,co))\<inverse> = cnf (X,rf,co)"
using com_cnf[OF assms] cnf_com[OF assms] by blast
  
end
  