theory no_txns_power imports no_txns begin
  
subsection "Model relations"
  
(* fences induced by transaction boundaries *)
fun tfence :: "witnessed_execution \<Rightarrow> event rel"
  where "tfence (X,rf,co) = po X \<inter> (into X (stxn X) \<union> outof X (stxn X))"
   
(* propagation induced by transactions (mode 1) *)
fun tprop1 :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> event rel"
  where "tprop1 ppo (X,rf,co) = fre (X,rf,co) O stxn X O rfe (X,rf,co) O ppo (X,rf,co)"
    
(* propagation induced by transactions (mode 2) *)
fun tprop2 :: "witnessed_execution \<Rightarrow> event rel"
  where "tprop2 (X,rf,co) = rfe (X,rf,co) O stxn X O \<lfloor> W X \<rfloor>"
    
(* propagation induced by transactions (mode 3) *)
fun tprop3 :: "witnessed_execution \<Rightarrow> event rel"
  where "tprop3 (X,rf,co) = stxn X O rfe (X,rf,co)"
    
subsection "Model axioms"
  
fun Coherence :: "witnessed_execution \<Rightarrow> bool"
  where "Coherence (X,rf,co) = acyclic ((po X \<inter> sloc X) \<union> com (X,rf,co))"
    
fun Order :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> bool"
  where "Order hb (X,rf,co) = acyclic (hb (X,rf,co))"
    
fun Propagation :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> bool"
  where "Propagation prop (X,rf,co) = acyclic (co \<union> prop (X,rf,co))"
    
fun StrongIsol :: "witnessed_execution \<Rightarrow> bool"
  where "StrongIsol (X,rf,co) = acyclic (stronglift (com (X,rf,co), stxn X))"
    
fun TxnOrder :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> bool"
  where "TxnOrder hb (X,rf,co) = acyclic(stronglift(hb (X,rf,co), stxn X))"
    
fun TxnPropagation :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> bool"
  where "TxnPropagation prop (X,rf,co) = acyclic(stronglift(co \<union> prop (X,rf,co), stxn X))"

fun TxnCancelsRMW :: "witnessed_execution \<Rightarrow> bool"
  where "TxnCancelsRMW (X,rf,co) = (rmw X \<inter> (tfence (X,rf,co))^* = {})"
    
subsection "Proof that our changes do not affect non-transactional programs"
    
(* The tfence relation is empty in the absence of transactions *)
theorem no_txns_power_part1:
  assumes "stxn X = {}"
  shows "tfence (X,rf,co) = {}"
  apply (unfold tfence.simps into_def outof_def)
  apply (simp add: assms)
  done
    
(* The tprop1, tprop2, and tprop3 relations are all empty in the absence of transactions *)
theorem no_txns_power_part2:
  fixes ppo -- "the definition of ppo is irrelevant"
  assumes "stxn X = {}"
  shows "tprop1 ppo (X,rf,co) = {}" and "tprop2 (X,rf,co) = {}" and "tprop3 (X,rf,co) = {}"
  by (simp add: assms)+
   
(* The StrongIsol axiom adds nothing in the absence of transactions *)
theorem no_txns_power_part3:
  assumes "stxn X = {}"
  shows "Coherence (X,rf,co) \<longrightarrow> StrongIsol (X,rf,co)"
proof
  assume "Coherence (X, rf, co)"
  hence "acyclic (com (X,rf,co))" 
    using Coherence.simps acyclic_subset by blast
  thus "StrongIsol (X, rf, co)" by (simp add: assms)
qed
    
(* The TxnOrder axiom adds nothing in the absence of transactions *)
theorem no_txns_power_part4:
  fixes hb -- "the definition of hb is irrelevant"
  assumes "stxn X = {}"
  shows "Order hb (X,rf,co) \<longrightarrow> TxnOrder hb (X,rf,co)"
  using no_txns_stronglift by (simp add: assms)
   
(* The TxnPropagation axiom adds nothing in the absence of transactions *)
theorem no_txns_power_part5:
  fixes "prop" -- "the definition of prop is irrelevant"
  assumes "stxn X = {}"
  shows "Propagation prop (X,rf,co) \<longrightarrow> TxnPropagation prop (X,rf,co)"
  using no_txns_stronglift by (simp add: assms)
 
(* The TxnCancelsRMW axiom adds nothing in the absence of transactions *)
theorem no_txns_power_part6:
  assumes "wf_exec (X,rf,co)"
  assumes "stxn X = {}"
  shows "TxnCancelsRMW (X,rf,co) = True"
proof (simp add: into_def outof_def del: tfence.simps)
  have "rmw X \<inter> Id = {}" 
    using assms(1) 
    apply -
    apply (erule wf_exec.cases, auto) 
    done
  thus "rmw X \<inter> (tfence (X, rf, co))\<^sup>* = {}" 
    using no_txns_power_part1[OF assms(2)] by simp
qed
    
end
  