theory no_txns_x86 imports no_txns begin
  
subsection "Model relations"
  
(* fences induced by transaction boundaries *)
fun tfence :: "witnessed_execution \<Rightarrow> event rel"
  where "tfence (X,rf,co) = po X \<inter> (into X (stxn X) \<union> outof X (stxn X))"
    
subsection "Model axioms"
  
fun Coherence :: "witnessed_execution \<Rightarrow> bool"
  where "Coherence (X,rf,co) = acyclic ((po X \<inter> sloc X) \<union> com (X,rf,co))"
    
fun Order :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> bool"
  where "Order hb (X,rf,co) = acyclic (hb (X,rf,co))"
    
fun StrongIsol :: "witnessed_execution \<Rightarrow> bool"
  where "StrongIsol (X,rf,co) = acyclic (stronglift (com (X,rf,co), stxn X))"
    
fun TxnOrder :: "(witnessed_execution \<Rightarrow> event rel) \<Rightarrow> witnessed_execution \<Rightarrow> bool"
  where "TxnOrder hb (X,rf,co) = acyclic(stronglift(hb (X,rf,co), stxn X))"
    
subsection "Proof that our changes do not affect non-transactional programs"
    
(* The tfence relation is empty in the absence of transactions *)
theorem no_txns_x86_part1:
  assumes "stxn X = {}"
  shows "tfence (X,rf,co) = {}"
  apply (unfold tfence.simps into_def outof_def)
  apply (simp add: assms)
  done
    
(* The StrongIsol axiom adds nothing in the absence of transactions *)
theorem no_txns_x86_part2:
  assumes "stxn X = {}"
  shows "Coherence (X,rf,co) \<longrightarrow> StrongIsol (X,rf,co)"
proof
  assume "Coherence (X, rf, co)"
  hence "acyclic (com (X,rf,co))" 
    using Coherence.simps acyclic_subset by blast
  thus "StrongIsol (X, rf, co)" by (simp add: assms)
qed
    
(* The TxnOrder axiom adds nothing in the absence of transactions *)
theorem no_txns_x86_part3:
  fixes hb -- "the definition of hb is irrelevant"
  assumes "stxn X = {}"
  shows "Order hb (X,rf,co) \<longrightarrow> TxnOrder hb (X,rf,co)"
  using no_txns_stronglift by (simp add: assms)
    
end
  